import re
import inspect
import numpy as np
import pandas as pd
from ..Utils import clean_keys


class DataProcessing:

    def __init__(self,
                 target=None,
                 num_cols=None,
                 date_cols=None,
                 cat_cols=None,
                 missing_values='interpolate',
                 outlier_removal='clip',
                 z_score_threshold=4,
                 folder='',
                 version=1,
                 mode='regression',
                 ):
        '''
        Preprocessing Class. Deals with Outliers, Missing Values, duplicate rows, data types (floats, categorical and
        dates), Not a Numbers, Infinities.

        :param target: Column name of target variable
        :param num_cols: Numerical columns, all parsed to integers and floats
        :param date_cols: Date columns, all parsed to pd.datetime format
        :param cat_cols: Categorical Columns. Currently all one-hot encoded.
        :param missing_values: How to deal with missing values ('remove', 'interpolate' or 'mean')
        :param outlier_removal: How to deal with outliers ('boxplot', 'zscore' or 'none')
        :param z_score_threshold: If outlierRemoval='zscore', the threshold is adaptable, default=4.
        :param folder: Directory for storing the output files
        :param version: Versioning the output files
        '''
        # Parameters
        self.folder = folder if len(folder) == 0 or folder[-1] == '/' else folder + '/'
        self.version = version
        self.target = re.sub('[^a-z0-9]', '_', target.lower())
        self.mode = mode
        self.numCols = [re.sub('[^a-z0-9]', '_', nc.lower()) for nc in num_cols] if num_cols is not None else []
        self.catCols = [re.sub('[^a-z0-9]', '_', cc.lower()) for cc in cat_cols] if cat_cols is not None else []
        self.dateCols = [re.sub('[^a-z0-9]', '_', dc.lower()) for dc in date_cols] if date_cols is not None else []
        if self.target in self.numCols:
            self.numCols.remove(self.target)

        # Variables
        self.scaler = None
        self.oScaler = None

        # Algorithms
        missing_values_implemented = ['remove_rows', 'remove_cols', 'interpolate', 'mean', 'zero']
        outlier_removal_implemented = ['boxplot', 'zscore', 'clip', 'none']
        if outlier_removal not in outlier_removal_implemented:
            raise ValueError(
                "Outlier Removal Algorithm not implemented. Should be in " + str(outlier_removal_implemented))
        if missing_values not in missing_values_implemented:
            raise ValueError("Missing Values Algorithm not implemented. Should be in " +
                             str(missing_values_implemented))
        self.missingValues = missing_values
        self.outlierRemoval = outlier_removal
        self.zScoreThreshold = z_score_threshold

    def clean(self, data):
        print('[Data] Data Cleaning Started, (%i x %i) samples' % (len(data), len(data.keys())))
        if len(data[self.target].unique()) == 2:
            self.mode = 'classification'

        # Clean
        data = clean_keys(data)
        data = self._remove_duplicates(data)
        data = self._convert_data_types(data)
        data = self._remove_outliers(data)
        data = self._remove_missing_values(data)
        data = self._remove_constants(data)

        # Finish
        self._store(data)
        print('[Data] Processing completed, (%i x %i) samples returned' % (len(data), len(data.keys())))
        return data

    def _convert_data_types(self, data):
        # Convert Data Types
        for key in self.dateCols:
            data.loc[:, key] = pd.to_datetime(data[key], errors='coerce', infer_datetime_format=True, utc=True)
        for key in [key for key in data.keys() if key not in self.dateCols and key not in self.catCols]:
            data.loc[:, key] = pd.to_numeric(data[key], errors='coerce', downcast='float')
        for key in self.catCols:
            if key in data.keys():
                dummies = pd.get_dummies(data[key])
                for dummy_key in dummies.keys():
                    dummies = dummies.rename(
                        columns={dummy_key: key + '_' + re.sub('[^a-z0-9]', '_', str(dummy_key).lower())})
                data = data.drop(key, axis=1).join(dummies)
        if self.target in data.keys():
            data.loc[:, self.target] = pd.to_numeric(data[self.target], errors='coerce')
        return data

    @staticmethod
    def _remove_duplicates(data):
        # Remove Duplicates
        data = data.drop_duplicates()
        data = data.loc[:, ~data.columns.duplicated()]
        return data

    @staticmethod
    def _remove_constants(data):
        # Remove Constants
        data = data.drop(columns=data.columns[data.nunique() == 1])
        return data

    def _remove_outliers(self, data):
        # Remove Outliers
        if self.outlierRemoval == 'boxplot':
            q1 = data.quantile(0.25)
            q3 = data.quantile(0.75)
            for key in q1.keys():
                data.loc[data[key] < q1[key] - 1.5 * (q3[key] - q1[key]), key] = np.nan
                data.loc[data[key] > q3[key] + 1.5 * (q3[key] - q1[key]), key] = np.nan
        elif self.outlierRemoval == 'zscore':
            z_score = (data - data.mean(skipna=True, numeric_only=True)) \
                     / data.std(skipna=True, numeric_only=True)
            data[z_score > self.zScoreThreshold] = np.nan
        elif self.outlierRemoval == 'clip':
            data = data.clip(lower=-1e12, upper=1e12)
        return data

    def _remove_missing_values(self, data):
        # Remove Missing Values
        data = data.replace([np.inf, -np.inf], np.nan)
        if self.missingValues == 'remove_rows':
            data = data[data.isna().sum(axis=1) == 0]
        elif self.missingValues == 'remove_cols':
            data = data.loc[:, data.isna().sum(axis=0) == 0]
        elif self.missingValues == 'zero':
            data = data.fillna(0)
        elif self.missingValues == 'interpolate':
            ik = np.setdiff1d(data.keys().to_list(), self.dateCols)
            data[ik] = data[ik].interpolate(limit_direction='both')
            if data.isna().sum().sum() != 0:
                data = data.fillna(0)
        elif self.missingValues == 'mean':
            data = data.fillna(data.mean())
        return data

    def _store(self, data):
        # Store cleaned data
        data.to_csv(self.folder + 'Cleaned_v%i.csv' % self.version, index_label='index')

    def export_function(self):
        function_strings = [
            inspect.getsource(clean_keys),
            inspect.getsource(self._remove_duplicates).replace('self.', ''),
            inspect.getsource(self._convert_data_types).replace('self.', ''),
            inspect.getsource(self._remove_outliers).replace('self.', ''),
            inspect.getsource(self._remove_missing_values).replace('self.', ''),
        ]
        function_strings = '\n'.join([k[k.find('\n'): k.rfind('\n', 0, k.rfind('\n'))] for k in function_strings])

        return """
            #################
            # Data Cleaning #
            #################
            # Copy vars
            catCols, dateCols, target = {}, {}, '{}'
            outlierRemoval, missingValues, zScoreThreshold = '{}', '{}', '{}'
    """.format(self.catCols, self.dateCols, self.target, self.outlierRemoval, self.missingValues,
               self.zScoreThreshold) + function_strings



class BaseClassifier:

    def __init__(self, model, params):
        self.defaultParams = params
        self.model = model(**params)
        self.hasPredictProba = hasattr(model, 'predict_proba')

    def get_params(self):
        return self.model.get_params()

    def set_params(self, args):
        self.model.set_params(**args)
        return self

    def predict(self, x):
        return self.model.predict(x)

    def predict_proba(self, x):
        if self.hasPredictProba:
            return self.model.predict_proba(x)

    def score(self, x, y):
        return self.model.score(x, y)

    def fit(self, x, y):
        self.model.fit(x, y)
        return self



import copy
import time
import optuna
import numpy as np
import pandas as pd
import multiprocessing as mp
import catboost
import xgboost
import lightgbm
from sklearn import svm
from sklearn import ensemble
from sklearn import linear_model
from .BaseGridSearch import BaseGridSearch


class OptunaGridSearch(BaseGridSearch):

    def __init__(self, model, params=None, cv=None, scoring=None, verbose=None, timeout=3600, n_trials=None):
        super().__init__(model, params, cv, scoring, verbose)
        self.timeout = timeout
        self.nTrials = n_trials
        self.x, self.y = None, None

    def get_params(self, trial):
        if isinstance(self.model, linear_model.Lasso) or \
           isinstance(self.model, linear_model.Ridge) or \
           isinstance(self.model, linear_model.RidgeClassifier):
            return {
                'alpha': trial.suggest_uniform('alpha', 0, 10),
            }
        elif isinstance(self.model, svm.SVC) or \
                isinstance(self.model, svm.SVR):
            return {
                'gamma': trial.suggest_categorical('gamma', ['scale', 'auto', 0.001, 0.01, 0.1, 0.5, 1]),
                'C': trial.suggest_uniform('C', 0, 10),
            }
        elif isinstance(self.model, catboost.CatBoostRegressor):
            return {
                    'loss_function': trial.suggest_categorical('loss', ['MAE', 'RMSE']),
                    'learning_rate': trial.suggest_uniform('learning_rate', 0, 1),
                    'l2_leaf_reg': trial.suggest_uniform('l2_leaf_reg', 0, 10),
                    'depth': trial.suggest_int('depth', 3, 15),
                    'min_data_in_leaf': trial.suggest_int('min_data_in_leaf', 1, 1000),
                    'max_leaves': trial.suggest_int('max_leaves', 10, 250),
                }
        elif isinstance(self.model, ensemble.GradientBoostingRegressor):
            return {
                'loss': trial.suggest_categorical('loss', ['ls', 'lad', 'huber']),
                'learning_rate': trial.suggest_uniform('learning_Rate', 0, 1),
                'max_depth': trial.suggest_int('max_depth', 3, 15),
            }
        elif isinstance(self.model, ensemble.HistGradientBoostingRegressor):
            return {
                'max_iter': trial.suggest_int('max_iter', 100, 250),
                'max_bins': trial.suggest_int('max_bins', 100, 255),
                'loss': trial.suggest_categorical('loss', ['least_squares', 'least_absolute_deviation']),
                'l2_regularization': trial.suggest_uniform('l2_regularization', 0, 10),
                'learning_rate': trial.suggest_uniform('learning_rate', 0, 1),
                'max_leaf_nodes': trial.suggest_int('max_leaf_nodes', 30, 150),
                'early_stopping': [True],
            }
        elif isinstance(self.model, ensemble.RandomForestRegressor):
            return {
                'criterion': trial.suggest_categorical('criterion', ['mse', 'mae']),
                'max_depth': trial.suggest_int('max_depth', 3, 15),
                'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt']),
                'min_samples_split': trial.suggest_int('min_samples_split', 2, 50),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, 1000),
                'bootstrap': trial.suggest_categorical('bootstrap', [True, False]),
            }
        elif isinstance(self.model, xgboost.XGBRegressor):
            return {
                'max_depth': trial.suggest_int('max_depth', 3, 15),
                'booster': trial.suggest_categorical('booster', ['gbtree', 'gblinear', 'dart']),
                'learning_rate': trial.suggest_uniform('learning_rate', 0, 10),
                'verbosity': [0],
                'n_jobs': [mp.cpu_count() - 1],
            }
        elif isinstance(self.model, lightgbm.LGBMRegressor):
            return {
                'num_leaves': trial.suggest_int('num_leaves', 10, 150),
                'min_child_samples': trial.suggest_int('min_child_samples', 1, 1000),
                'min_child_weight': trial.suggest_uniform('min_child_weight', 0, 1),
                'subsample': trial.suggest_uniform('subsample', 0, 1),
                'colsample_bytree': trial.suggest_uniform('colsample_bytree', 0, 1),
                'reg_alpha': trial.suggest_uniform('reg_alpha', 0, 1),
                'reg_lambda': trial.suggest_uniform('reg_lambda', 0, 1),
                'n_jobs': [mp.cpu_count() - 1],
            }
        elif isinstance(self.model, catboost.CatBoostClassifier):
            return {
                'loss_function': trial.suggest_categorical('loss_function', ['Logloss' if self.y.nunique() == 2 else
                                                                             'MultiClass']),
                'learning_rate': trial.suggest_uniform('learning_rate', 0, 1),
                'l2_leaf_reg': trial.suggest_uniform('l2_leaf_reg', 0, 10),
                'depth': trial.suggest_int('depth', 1, 10),
                'min_data_in_leaf': trial.suggest_randint('min_data_in_leaf', 50, 500),
                'grow_policy': trial.suggest_categorical('grow_policy', ['SymmetricTree', 'Depthwise', 'Lossguide']),
            }
        elif isinstance(self.model, ensemble.GradientBoostingClassifier):
            return {
                'loss': trial.suggest_categorical('loss', ['deviance', 'exponential']),
                'learning_rate': trial.suggest_uniform('learning_rate', 0, 1),
                'max_depth': trial.suggest_int('max_depth', 3, 15),
            }
        elif isinstance(self.model, ensemble.HistGradientBoostingClassifier):
            return {
                'max_iter': trial.suggest_int('max_iter', 100, 250),
                'max_bins': trial.suggest_int('max_bins', 100, 255),
                'l2_regularization': trial.suggest_uniform('l2_regularization', 0, 10),
                'learning_rate': trial.suggest_uniform('learning_Rate', 0, 1),
                'max_leaf_nodes': trial.suggest_int('max_leaf_nodes', 30, 150),
                'early_stopping': [True]
            }
        elif isinstance(self.model, ensemble.RandomForestClassifier):
            return {
                'criterion': trial.suggest_categorical('criterion', ['gini', 'entropy']),
                'max_depth': trial.suggest_int('max_depth', 3, 15),
                'max_features': trial.suggest_categorical('max_features', ['auto', 'sqrt']),
                'min_samples_split': trial.suggest_int('min_samples_split', 2, 50),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', 1, 1000),
                'bootstrap': trial.suggest_categorical('bootstrap', [True, False]),
            }
        elif isinstance(self.model, xgboost.XGBClassifier):
            return {
                'max_depth': trial.suggest_int('max_depth', 3, 15),
                'booster': trial.suggest_categorical('booster', ['gbtree', 'gblinear', 'dart']),
                'learning_rate': trial.suggest_uniform('learning_rate', 0, 10),
                'verbosity': [0],
                'n_jobs': [mp.cpu_count() - 1],
                'scale_pos_weight': trial.suggest_uniform('scale_pos_weight', 0, 100)
            }
        elif isinstance(self.model, lightgbm.LGBMClassifier):
            return {
                'num_leaves': trial.suggest_int('num_leaves', 10, 150),
                'min_child_samples': trial.suggest_int('min_child_samples', 1, 1000),
                'min_child_weight': trial.suggest_uniform('min_child_weight', 0, 1),
                'subsample': trial.suggest_uniform('subsample', 0, 1),
                'colsample_bytree': trial.suggest_uniform('colsample_bytree', 0, 1),
                'reg_alpha': trial.suggest_uniform('reg_alpha', 0, 1),
                'reg_lambda': trial.suggest_uniform('reg_lambda', 0, 1),
                'n_jobs': [mp.cpu_count() - 1],
            }

        # Raise error if nothing is returned
        raise NotImplementedError('Hyper parameter tuning not implemented for ', type(self.model).__name__)

    def get_callbacks(self):
        pass

    def fit(self, x, y):
        # Store
        self.x, self.y = x, y
        # Set up study
        study = optuna.create_study(sampler=optuna.samplers.TPESampler())
        study.optimize(self.objective, timeout=self.timeout, n_trials=self.nTrials, show_progress_bar=True)

        # Parse results
        optuna_results = study.trials_dataframe()
        results = pd.DataFrame({
            'worst_case': optuna_results['value'] - optuna_results['user_attrs_std_value'],
            'mean_objective': optuna_results['value'],
            'std_objective': optuna_results['user_attrs_std_value'],
            'params': [x.params for x in study.get_trials()],
            'mean_time': optuna_results['user_attrs_mean_time'],
            'std_time': optuna_results['user_attrs_std_time']
        })
        return results

    def objective(self, trial):
        # Metrics
        scores = []
        times = []

        # Cross Validation
        for t, v in self.cv.split(self.x, self.y):
            # Split data
            xt, xv, yt, yv = self.x.iloc[t], self.x.iloc[v], self.y.iloc[t], self.y.iloc[v]

            # Train model
            t_start = time.time()
            model = copy.copy(self.model)
            model.set_params(**self.get_params(trial))
            model.fit(xt, yt)

            # Results
            scores.append(self.scoring(model, xv, yv))
            times.append(time.time() - t_start)

        trial.set_user_attr('mean_time', np.mean(times))
        trial.set_user_attr('std_time', np.std(times))
        trial.set_user_attr('std_value', np.std(scores))
        return np.mean(scores)



import os
import copy
import time
import joblib
import datetime
import numpy as np
import pandas as pd
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold
from ..Classifiers.CatBoostClassifier import CatBoostClassifier
from sklearn import linear_model
from sklearn import ensemble
from sklearn import svm


class Modelling:

    def __init__(self,
                 mode='regression',
                 shuffle=False,
                 scoring=None,
                 folder='models/',
                 n_splits=3,
                 dataset=0,
                 store_models=False,
                 store_results=True):
        """
        Class that analyses the performance of various regressors or classifiers.
        @param [str] mode: 'regression' or 'classification'
        @param [bool] shuffle: Whether to shuffle samples for training / validation
        @param [SciKit make_scorer] scoring: Performance metric
        @param [str] folder: Folder to store models and / or results
        @param [int] n_splits: Number of cross-validation splits
        @param [str] dataset: Name of feature set to use
        @param [bool] store_models: Whether to store the trained models
        @param [bool] store_results: Whether to store the results
        """
        # Args
        self.scoring = scoring
        self.mode = mode
        self.shuffle = shuffle
        self.cvSplits = n_splits
        self.dataset = str(dataset)
        self.storeResults = store_results
        self.storeModels = store_models
        self.folder = folder if folder[-1] == '/' else folder + '/'

        self.samples = None      # Number of samples in data
        self.needsProba = False  # Whether scorer requires needs_proba attr
        if 'True' in self.scoring._factory_args():
            self.needsProba = True

    def fit(self, x, y):
        # Copy number of samples
        self.samples = len(y)

        # Regression
        if self.mode == 'regression':
            cv = KFold(n_splits=self.cvSplits, shuffle=self.shuffle)
            return self._fit(x, y, cv)

        # Classification
        if self.mode == 'classification':
            cv = StratifiedKFold(n_splits=self.cvSplits, shuffle=self.shuffle)
            return self._fit(x, y, cv)

    def return_models(self):
        models = []

        # All classifiers
        if self.mode == 'classification':
            # The thorough ones
            if self.samples < 25000:
                if not self.needsProba:     # Arent probabilistic
                    models.append(linear_model.RidgeClassifier())
                    models.append(svm.SVC(kernel='rbf'))
                models.append(ensemble.BaggingClassifier())
                models.append(ensemble.GradientBoostingClassifier())
                # todo XG Boost

            # The efficient ones
            else:
                if not self.needsProba:
                    models.append(linear_model.RidgeClassifier())
                models.append(ensemble.HistGradientBoostingClassifier())
                # todo LGBM

            # And the multifaceted ones
            models.append(CatBoostClassifier())
            models.append(ensemble.RandomForestClassifier())

        elif self.mode == 'regression':
            # The thorough ones
            if self.samples < 25000:
                if not self.needsProba:     # Arent probabilistic
                    models.append(linear_model.LinearRegression())
                    models.append(svm.SVR(kernel='rbf'))
                models.append(ensemble.BaggingRegressor())
                models.append(ensemble.GradientBoostingRegressor())
                # todo XG Boost

            # The efficient ones
            else:
                if not self.needsProba:
                    models.append(linear_model.LinearRegression())
                models.append(ensemble.HistGradientBoostingRegressor())
                # todo LGBM

            # And the multifaceted ones
            # todo Catboost
            # models.append(CatBoostClassifier())
            models.append(ensemble.RandomForestRegressor())

        # Filter predict_proba models
        models = [m for m in models if hasattr(m, 'predict_proba')]

        return models

    def _fit(self, x, y, cross_val):
        # Convert to NumPy
        x = np.array(x)
        y = np.array(y).ravel()

        # Data
        print('[Modelling] Splitting data (shuffle=%s, splits=%i, features=%i)' % (str(self.shuffle), self.cvSplits, len(x[0])))

        if self.store_results and 'Initial_Models.csv' in os.listdir(self.folder):
            results = pd.read_csv(self.folder + 'Initial_Models.csv')
        else:
            results = pd.DataFrame(columns=['date', 'model', 'dataset', 'params', 'mean_objective', 'std_objective', 'mean_time', 'std_time'])

        # Models
        self.models = self.return_models()

        # Loop through models
        for master_model in self.models:
            # Check first if we don't already have the results
            ind = np.where(np.logical_and(np.logical_and(
                results['model'] == type(master_model).__name__,
                results['dataset'] == self.dataset),
                results['date'] == datetime.today().strftime('%d %b %y, %Hh')))[0]
            if len(ind) != 0:
                print('[Modelling] %s %s: %.4f, training time: %.1f s' %
                      (type(model).__name__.ljust(60), self.scoring._score_func.__name__,
                       np.mean(results.iloc[0]['mean_objective']), time.time() - t_start))
                continue

            # Time & loops through Cross-Validation
            val_score = []
            train_score = []
            train_time = []
            for t, v in cross_val.split(x, y):
                t_start = time.time()
                xt, xv, yt, yv = x[t], x[v], y[t], y[v]
                model = copy.copy(master_model)
                model.fit(xt, yt)
                val_score.append(self.scoring(model, xv, yv))
                train_score.append(self.scoring(model, xt, yt))
                train_time.append(time.time() - t_start)

            # Results
            results = results.append({'date': datetime.today().strftime('%d %b %y'), 'model': type(model).__name__,
                                      'dataset': self.dataset, 'params': model.get_params(),
                                      'mean_objective': np.mean(val_score),
                                      'std_objective': np.std(val_score), 'mean_time': np.mean(train_time),
                                      'std_time': np.std(train_time)}, ignore_index=True)

            print('[Modelling] %s %s: %.4f, training time: %.1f s' %
                  (type(model).__name__.ljust(60), self.scoring._score_func.__name__,
                   np.mean(results.iloc[0]['mean_objective']), time.time() - t_start))

            if self.store_models:
                joblib.dump(model, self.folder + type(model).__name__ + '_%.4f.joblib' % self.acc[-1])

        # Store CSV
        if self.store_results:
            results.to_csv(self.folder + 'Initial_Models.csv')

        return results
